import logging
from typing import Literal, Optional

import requests
from requests.exceptions import ChunkedEncodingError, ConnectionError, HTTPError

from sdci.exceptions import SDCIException
from sdci.schemas import TaskOutputSchema
from sdci.settings import CLIENT_REQUEST_TIMEOUT_SECONDS

logger = logging.getLogger(__name__)


class SDCIClient:
    def __init__(self, endpoint, token) -> None:
        self._client = requests.Session()
        self._endpoint = endpoint
        self._token = token

    def trigger(
        self, task_name, *args, action: Literal["run", "status"] = "run"
    ) -> Optional[TaskOutputSchema]:
        match action:
            case "run":
                action_url = f"{self._endpoint}/tasks/{task_name}/"
            case "status":
                action_url = f"{self._endpoint}/tasks/{task_name}/status/"

        try:
            response = self._client.post(
                action_url,
                headers={
                    "Authorization": f"Bearer {self._token}",
                },
                json={
                    "args": args,
                }
                if action == "run"
                else {},
                stream=action == "run",
                timeout=CLIENT_REQUEST_TIMEOUT_SECONDS,
            )
        except HTTPError as exc:
            raise SDCIException(f"CLIENT HTTPError: {exc}") from exc

        except ChunkedEncodingError as exc:
            raise SDCIException(
                f"CLIENT CHUNKED ENCODING ERROR (POTENTIALLY SERVER ERROR): {exc}"
            ) from exc

        except ConnectionError as exc:
            raise SDCIException(f"SERVER UNAVAILABLE\n\n{exc}") from exc

        if response.status_code == 401:
            raise SDCIException("SERVER UNAUTHORIZED - Please check token")

        if response.status_code == 422:
            raise SDCIException(
                f"SERVER TASK UNAVAILABLE: {response.json().get('detail', 'Unknown Error')}"
            )

        if response.status_code == 429:
            raise SDCIException("SERVER HAS A WORKING TASK - PLEASE WAIT")

        if action == "status":
            return TaskOutputSchema.model_validate(response.json())

        if action == "run":
            with response:
                for line in response.iter_lines():
                    line = line.decode("utf-8")
                    print(line)



import os
from importlib.metadata import version

import click

__version__ = version("sdci")

from sdci.client_service import SDCIClient
from sdci.exceptions import SDCIException


@click.group()
def entrypoint():
    print(f"[ SDCI-CLI v{__version__} ]")
    """CLI application"""
    pass


@entrypoint.command()
@click.option("--token", required=False, help="Token or env var SDCI_TOKEN")
@click.argument("server", required=False)
@click.argument("task", required=True)
@click.argument("args", nargs=-1)
def run(token, server, task, args):
    """Run a task into the server"""
    print(f"[ TASK RUN ] - Running {task} into server {server}...\n")

    if not token:
        token = os.environ.get("SDCI_TOKEN", None)

    try:
        if not token:
            raise SDCIException(
                "TOKEN NOT FOUND - Please provide a token or set SDCI_TOKEN env var"
            )

        client = SDCIClient(server, token)
        client.trigger(task, *args, action="run")
        output = client.trigger(task, ..., action="status")
        exit(output.exit_code)

    except SDCIException as exc:
        print(f"[Client Failed to execute task] - {exc}")
        exit(1)



from typing import Literal, Optional

from pydantic import BaseModel


class TaskRequestSchema(BaseModel):
    args: list = []


class TaskOutputSchema(BaseModel):
    pid: Optional[int] = None
    exit_code: Optional[int] = None
    status: Literal["STOPPED", "RUNNING", "FINISHED", "TIMEOUT"] = "STOPPED"



import asyncio
import logging
import os
import time
from importlib.metadata import version

from sdci.exceptions import SDCIServerException
from sdci.settings import TASK_RUN_TIMEOUT_SECONDS

__version__ = version("sdci")
logger = logging.getLogger(__name__)


class CommandRunner:
    def __init__(self, shell_file: str) -> None:
        self._task_name = shell_file
        self._shell_file = f"./tasks/{self._task_name}.sh"
        self._lock = None
        self._store = None

        if not os.path.exists(self._shell_file):
            raise SDCIServerException(
                f"SHELL FILE NOT FOUND ON SERVER: {self._shell_file}"
            )

    def for_lock(self, lock: asyncio.Lock):
        self._lock = lock
        return self

    def for_store(self, store: dict):
        self._store = store
        return self

    async def run(self, args):
        if not self._lock:
            raise SDCIServerException("NO LOCK available")

        cmd = ["bash", self._shell_file]
        cmd.extend(args)

        logger.info(f"RUNNING TASK WITH CMD: {cmd}")

        await self._lock.acquire()
        logger.info("Triggering Task - Lock acquired")

        yield f"********** SDCI SERVER v{__version__} **********\n"
        yield f'RUNNING TASK: "{cmd}": TIMEOUT: {TASK_RUN_TIMEOUT_SECONDS}s\n'
        yield "********** task logs - start ********** \n\n"

        process = await asyncio.create_subprocess_exec(
            *cmd, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
        )

        # Update Process
        run_status = "RUNNING"
        self._store[self._task_name] = {
            "pid": process.pid,
            "exit_code": None,
            "status": run_status,
        }

        start_time = time.time()
        timeout = start_time + TASK_RUN_TIMEOUT_SECONDS
        while True:
            output = await process.stdout.readline()

            if output:
                yield output.decode()

            if process.returncode is not None:
                run_status = "FINISHED"
                break

            if time.time() > timeout:
                process.kill()
                run_status = "TIMEOUT"
                yield "TIMEOUT REACHED"
                break

        self._store[self._task_name] = {
            "pid": process.pid,
            "exit_code": process.returncode,
            "status": run_status,
        }

        self._lock.release()
        logger.info("TASK ENDED - Lock released")

        yield "\n********** task logs - end ********** \n"

        elapsed_time = time.time() - start_time
        yield f"EXITED ({process.returncode}) - Took {elapsed_time:.2f}s"



# SDCI Settings
import os

SERVER_TOKEN = os.environ.get("SDCI_SERVER_TOKEN", None)
TASK_RUN_TIMEOUT_SECONDS = os.environ.get("TASK_RUN_TIMEOUT", 120)
CLIENT_REQUEST_TIMEOUT_SECONDS = os.environ.get("CLIENT_REQUEST_TIMEOUT", 600)
